/*********************************************************************
  Weather station with a Nextion display, showing date, time, tempe-
  rature, humidity, airpressure and weather forecast using
  openweathermap.org

  Setup used:
  - NodeMCU (ESP8266)
  - Nextion 7" 800x480 display

**********************************************************************
  Last updated 20200128 by McUles

  This code is free for personal use, not for commercial purposes.
  Please leave this header intact.

  contact: dev@itstall.de
**********************************************************************/
//#define DEBUG
#define DEBUG_NO_PAGE_FADE
#define DEBUG_NO_TOUCH
#define nexSerial Serial1

//
// include libraries
//
#include <Wire.h>
#include <SPI.h>                                // I2C library
#include <ArduinoJson.h>                        // https://github.com/bblanchon/ArduinoJson
#include <ESP8266WiFi.h>                        // WiFi library
#include <WiFiUdp.h>                            // Udp library
#include <TimeLib.h>                            // Time library
#include "NTPClient.h"
#include "PubSubClient.h"                       // http://knolleary.net/arduino-client-for-mqtt/
#include "settings.h"

extern "C" {
  #include "user_interface.h"
}

unsigned int localPort = 2390;                  // local port to listen for UDP packets
IPAddress timeServerIP;                         // IP address of random server
IPAddress mqttHostIP;                           // IP address of random server
byte packetBuffer[48];                          // buffer to hold incoming and outgoing packets

//
// initialize variables
//
int page = 1;
String command;
String doubleQuote = "\"\"";

WiFiClient wifiClient;
WiFiUDP ntpUDP;
NTPClient timeClient(ntpUDP, ntpServerName, timeZoneoffsetGMT, timeServerRefreshDelay);
PubSubClient mqttClient(wifiClient);

void setup() {
//#ifdef DEBUG
  Serial.begin(9600);
  Serial.println("setup...");
//#endif

  nexSerial.begin(9600);

  printNextionCommand("dims=" + String(0));         // set initial startup backlight value of the Nextion display to 0
  printNextionCommand("dim=" + String(0));          // also set the current backlight value to 0
  printNextionCommand("page 1");

  delay(startupDelay);

  displayFadeIn(1, displayStartupDimValue, dimStartupDelay);

  reconnectToWifi();
  timeClient.begin();
  getTimeFromServer();

  displayFadeIn(displayStartupDimValue, displayDimValue, dimStartupDelay / 2);

  mqttClient.setServer(mqttHost, mqttPort);
  mqttClient.setCallback(getWeatherDataMqtt);
  reconnectMqtt();
}
void loop() {
#ifdef DEBUG
  Serial.println("loop");
#endif

  if (page == 0) {

    if (weatherForecastLoop == maxForecastLoop) {
      page = 1;
      weatherForecastLoopInc = -weatherForecastLoopInc;
    }

    delayCheckTouch(loopDelay);
  }

  else if (page == 1) {
    if (weatherForecastLoop == 0) {
      //page = 0;

#if !defined (DEBUG_NO_PAGE_FADE)
      displayFadeOut(displayDimValue, dimPageDelay);
#endif

      //printNextionCommand("page 0");

#if !defined (DEBUG_NO_PAGE_FADE)
      displayFadeIn(0, displayDimValue, dimPageDelay);
#endif

      weatherForecastLoopInc = -weatherForecastLoopInc;
    }
    else
      delayCheckTouch(loopDelay);
  }

  weatherForecastLoop += weatherForecastLoopInc;

  // Update date and time on display
  sendToLCD(1, "time", String(timeClient.getFormattedTime()));
  sendToLCD(1, "dateDay", dayAsString(timeClient.getDay()));
  sendToLCD(1, "date", String(timeClient.getFormattedDate()));

  reconnectToWifi();

  reconnectMqtt();
  mqttClient.loop();

#ifdef DEBUG
    Serial.print(page);
    Serial.print(" ");
    Serial.print(weatherForecastLoop);
    Serial.println("");
#endif
}
void printNextionCommand (String command) {
  nexSerial.print(command);
  endNextionCommand();
}
void sendToLCD(uint8_t type, String index, String cmd) {
#ifdef DEBUG
  Serial.println("sendToLCD");
#endif
  if (type == 1 ) {
    nexSerial.print(index);
    nexSerial.print(".txt=");
    nexSerial.print("\"");
    nexSerial.print(cmd);
    nexSerial.print("\"");
  }
  else if (type == 2) {
    nexSerial.print(index);
    nexSerial.print(".val=");
    nexSerial.print(cmd);
  }
  else if (type == 3) {
    nexSerial.print(index);
    nexSerial.print(".pic=");
    nexSerial.print(cmd);
  }
  else if (type == 4 ) {
    nexSerial.print("page ");
    nexSerial.print(cmd);
  }

  endNextionCommand();
}
void endNextionCommand() {
  nexSerial.write(0xff);
  nexSerial.write(0xff);
  nexSerial.write(0xff);
}
void displayFadeIn(int fromValue, int toValue, int fadeDelay) {
#ifdef DEBUG
  Serial.println("displayFadeIn");
#endif
  for (int i = fromValue; i <= toValue; i += displayDimStep) {
    if (i > toValue) {
      i = toValue;
    }
    printNextionCommand("dim=" + String(i));
    delay(fadeDelay);
  }
}
void displayFadeOut(int fromValue, int fadeDelay) {
#ifdef DEBUG
  Serial.println("displayFadeOut");
#endif
  for (int i = fromValue; i >= 0; i -= displayDimStep) {
    if (i < 0) {
      i = 0;
    }
    printNextionCommand("dim=" + String(i));
    delay(fadeDelay);
  }
}
void reconnectToWifi() {
#ifdef DEBUG
  Serial.println("reconnectToWifi...");
#endif
  if(WiFi.status() != WL_CONNECTED) {
    int wifiBlink = 0;
  
    // WiFi.enableSTA(true);
    WiFi.mode(WIFI_STA);
    WiFi.disconnect();
    WiFi.begin(ssid, password);
  
    while (WiFi.status() != WL_CONNECTED) {
      if (wifiBlink == 0) {
        sendToLCD(1, "txt_ip", "\"No IP yet...\"");
        sendToLCD(3, "btn_wifi", "31");
        wifiBlink = 1;
      }
      else {
        //printNextionCommand("wifi_connect.txt=" + doubleQuote);
        sendToLCD(1, "txt_ip", doubleQuote);
        sendToLCD(3, "btn_wifi", "30");
        wifiBlink = 0;
      }
      delay(500);
    }
  }
#ifdef DEBUG
    Serial.print("Wifi IP Address: ");
    Serial.println(WiFi.localIP().toString());
#endif
  sendToLCD(1, "txt_ip", WiFi.localIP().toString());
  sendToLCD(3, "btn_wifi", "30");
  //printNextionCommand("wifi_connect.txt=" + doubleQuote);
}
void getTimeFromServer() {
#ifdef DEBUG
  Serial.println("getTimeFromServer");
#endif

  sendToLCD(1, "time", "get time..");
  while(!timeClient.update()) {
    timeClient.forceUpdate();
  }
  timeClient.setTimeOffset(timeZoneoffsetGMT);
  timeClient.setUpdateInterval(timeServerRefreshDelay);
}
void reconnectMqtt() {
  String settings;
  StaticJsonDocument<1024> doc;
  JsonObject object = doc.to<JsonObject>();
  object["plz"] = backendPlz;
  object["lngCode"] = backendLngCode;
  object["frontendId"] = frontendId;
  serializeJson(doc, settings);

  while (!mqttClient.connected()) {
    sendToLCD(3, "btn_mqtt", "28");
    if (mqttClient.connect(frontendId.c_str(), mqttUser, mqttPass)) {
      Serial.println("MQTT Connected");
      sendToLCD(3, "btn_mqtt", "29");
      mqttClient.subscribe(("/wetterstation/frontend/" + frontendId + "/weather").c_str());
      mqttClient.publish("/wetterstation/backend/settings", settings.c_str());
    }
    else {
#ifdef DEBUG
      Serial.print("failed, rc=");
      Serial.print(mqttClient.state());
      Serial.println(" retrying in 5 seconds");
#endif
      delay(5000);
    }
  }
  if(mqttClient.connected()) {
    sendToLCD(3, "btn_mqtt", "29");
  }
}
void getWeatherDataMqtt(char* topic, byte* payload, unsigned int length) {
#ifdef DEBUG
  Serial.println("getWeatherDataMqtt...");
  Serial.println("From topic: " + String(topic));
#endif
  // Parsing
  StaticJsonDocument<1024> root;
  deserializeJson(root, payload, length);

  String tmp0 = backendCity;
  float  tmp1 = root["temp"];          // Icon main
  float  tmp3 = root["tempMin"];
  float  tmp4 = root["tempMax"];
  float  tmp5 = root["humidity"];
  float  tmp6 = root["clouds"];
  float  tmp7 = root["rain3h"];
  float  tmp8 = root["snow3h"];
  float  tmp9 = root["windSpeed"];
  int    tmp10 = root["windDeg"];
  float  tmp11 = root["pressure"];

#if !defined (DEBUG_NO_PAGE_FADE)
  displayFadeOut(displayDimValue, dimPageDelay);
#endif

  printNextionCommand("page 1");

#if !defined (DEBUG_NO_PAGE_FADE)
  displayFadeIn(0, displayDimValue, dimPageDelay );
#endif

  setWeatherPicture(root["icon1Id"]);
  sendToLCD(1, "city", tmp0);
  //sendToLCD(1, "description", tmp2);
  sendToLCD(1, "humidity", String(tmp5, 0));
  sendToLCD(1, "rain", String(tmp7, 1));
  sendToLCD(1, "snow", String(tmp8, 1));
  sendToLCD(1, "wind_dir", getShortWindDirection(tmp10));
  sendToLCD(1, "wind_speed", String(tmp9, 1));
  sendToLCD(1, "pressure", String(tmp11, 0));
  sendToLCD(1, "clouds", String(tmp6, 0));
  sendToLCD(1, "temp_min", String(tmp3, 1));
  sendToLCD(1, "temp_max", String(tmp4, 1));
  sendToLCD(1, "temp", String(tmp1, 1));
}
String getWindDirection (int degrees) {
#ifdef DEBUG
  Serial.println("getWindDirection...");
#endif
  int sector = ((degrees + 11) / 22.5 - 1);
  switch (sector) {
    case 0: return "Nord";
    case 1: return "Nord-Nordost";
    case 2: return "Nordost";
    case 3: return "Ost-Nordost";
    case 4: return "Ost";
    case 5: return "Ost-Südost";
    case 6: return "Südost";
    case 7: return "Süd-Südost";
    case 8: return "Süd";
    case 9: return "Süd-Südwest";
    case 10: return "Südwest";
    case 11: return "West-Nordwest";
    case 12: return "West";
    case 13: return "West-Nordwest";
    case 14: return "Nordwest";
    case 15: return "Nord-Nordwest";
  }
}
String getShortWindDirection (int degrees) {
#ifdef DEBUG
  Serial.println("getShortWindDirection...");
#endif
  int sector = ((degrees + 11) / 22.5 - 1);
  switch (sector) {
    case 0: return "N";
    case 1: return "NNO";
    case 2: return "NO";
    case 3: return "ONO";
    case 4: return "O";
    case 5: return "OSO";
    case 6: return "SO";
    case 7: return "SSO";
    case 8: return "S";
    case 9: return "SSW";
    case 10: return "SW";
    case 11: return "WSW";
    case 12: return "W";
    case 13: return "WNW";
    case 14: return "NW";
    case 15: return "NNW";
  }
}
void setWeatherPicture(int weatherID) {
#ifdef DEBUG
  Serial.println("setWeatherPicture...");
#endif
  switch (weatherID) {
    case 200:
    case 201:
    case 202:
    case 210: sendToLCD(3, "weatherpic", "26"); break; // tstorm1
    case 211: sendToLCD(3, "weatherpic", "27"); break; // tstorm2
    case 212: sendToLCD(3, "weatherpic", "28"); break; // tstorm3
    case 221:
    case 230:
    case 231:
    case 232: sendToLCD(3, "weatherpic", "27"); break; // tstorm2

    case 300:
    case 301:
    case 302:
    case 310:
    case 311:
    case 312:
    case 313:
    case 314:
    case 321: sendToLCD(3, "weatherpic", "15"); break; // rain1

    case 500:
    case 501: sendToLCD(3, "weatherpic", "15"); break; // rain1
    case 502:
    case 503:
    case 504: sendToLCD(3, "weatherpic", "16"); break; // rain2
    case 511:
    case 520:
    case 521: sendToLCD(3, "weatherpic", "17"); break; // shower1
    case 522:
    case 531: sendToLCD(3, "weatherpic", "18"); break; // shower2

    case 600: sendToLCD(3, "weatherpic", "20"); break; // snow1
    case 601: sendToLCD(3, "weatherpic", "22"); break; // snow3
    case 602: sendToLCD(3, "weatherpic", "24"); break; // snow5
    case 611:
    case 612: sendToLCD(3, "weatherpic", "14"); break; // sleet
    case 615: sendToLCD(3, "weatherpic", "20"); break; // snow1
    case 616: sendToLCD(3, "weatherpic", "22"); break; // snow3
    case 620: sendToLCD(3, "weatherpic", "20"); break; // snow1
    case 621: sendToLCD(3, "weatherpic", "22"); break; // snow3
    case 622: sendToLCD(3, "weatherpic", "24"); break; // snow5

    case 701:
    case 711:
    case 721: sendToLCD(3, "weatherpic", "13"); break; // mist
    case 731: sendToLCD(3, "weatherpic", "10"); break; // dunno
    case 741: sendToLCD(3, "weatherpic", "11"); break; // fog
    case 751:
    case 761:
    case 762:
    case 771:
    case 781: sendToLCD(3, "weatherpic", "10"); break; // dunno

    case 800: sendToLCD(3, "weatherpic", "25"); break; // sunny
    case 801: sendToLCD(3, "weatherpic", "5"); break; // cloud1
    case 802: sendToLCD(3, "weatherpic", "7"); break; // cloud3
    case 803: sendToLCD(3, "weatherpic", "8"); break; // cloud4
    case 804: sendToLCD(3, "weatherpic", "14"); break; // overcast

    default: sendToLCD(3, "weatherpic", "10"); break; // dunno
  }
}
void delayCheckTouch (int delayTime) {
#ifdef DEBUG
  Serial.println("delayCheckTouch...");
#endif
  unsigned long startMillis = millis();

  while (millis() - startMillis < delayTime) {
    delay(1000);
  }
}
String dayAsString(int day) {
#ifdef DEBUG
  Serial.println("dayAsString...");
#endif
  switch (day) {
    case 1: return "Montag";
    case 2: return "Dienstag";
    case 3: return "Mittwoch";
    case 4: return "Donnerstag";
    case 5: return "Freitag";
    case 6: return "Samstag";
    case 7: return "Sonntag";
  }
  return "" ;
}
String monthAsString(int month) {
#ifdef DEBUG
  Serial.println("monthAsString...");
#endif
  switch (month) {
    case 1:  return "Januar";
    case 2:  return "Februar";
    case 3:  return "März";
    case 4:  return "April";
    case 5:  return "Mai";
    case 6:  return "Juni";
    case 7:  return "Juli";
    case 8:  return "August";
    case 9:  return "September";
    case 10: return "Oktober";
    case 11: return "November";
    case 12: return "Dezember";
  }
  return "" ;
}
